/*globals MXD*/

var V = MXD.Vector;

var allAxes = [
  [-1, 0],
  [ 1, 0],
  [0, -1],
  [0,  1]
];

function rangeUnion(range, x) {
  if (range.length == 0)
    return [x, x];
  else if (x < range[0])
    return [x, range[1]];
  else if (x > range[1])
    return [range[0], x];
  else
    return range;
}

function projectPoint(axis, p) {
  return V.dot(p, axis);
}

function projectRect(axis, rect) {
  var range = [];
  rect.corners.forEach(function(corner) {
    range = rangeUnion(range, projectPoint(axis, corner));
  });
  return range;
}

function rangeIntersect(b, a) {
  if (b[0] > a[1])
    return null;

  if (b[1] < a[0])
    return null;

  var ab = a[1] - b[0];
  var ba = a[0] - b[1];

  return (Math.abs(ab) < Math.abs(ba)) ? ab : ba;
}

function collideRectRect(a, b) {

  var bestAxis = null;
  var best = null;
  var bestr;


  var axes = a.axes || b.axes || allAxes;

  for (var axis of axes) {
    var ra = projectRect(axis, a);
    var rb = projectRect(axis, b);
    var inter = rangeIntersect(rb, ra);
    if (inter === null)
      return null;
    if (best == null || Math.abs(inter) < Math.abs(best)) {
      bestr = [ra, rb];
      best = inter;
      bestAxis = axis;
    }
  }

  if (!best)
    return null;

  return {
    axis:  bestAxis,
    depth: best
  };
}
