/*globals MXD, _, collideRectRect*/

var Rect = MXD.Rect;
var V = MXD.Vector;

var canvas = document.createElement('canvas');
document.body.appendChild(canvas);
document.write('<br>A=left D=right J=jump K=attack');
canvas.style.imageRendering = 'pixelated';

canvas.style.width = '100%';
//canvas.style.height = (canvas.height * 2) + 'px';
window.onresize = function() {
}
canvas.style.border = '1px solid black';

var context = canvas.getContext('2d');

var preAttackTime = 0.1;
var attackTime = 0.25;
var gravity = [0, 300];
var walkSpeed = 80;
var friction = 5;
var jumpHeight = 35;
var jumpVel = Math.sqrt(2 * gravity[1] * jumpHeight);
var floor = 150;

var paused = false;
window.onfocus = function() { paused = false; };
window.onblur = function() { paused = true; };

window.onkeydown = function(e) { key(e, 'down'); };
window.onkeyup = function(e) { key(e, 'up'); };

var frameNo = 0;
var keys = {};
var intent = [0, 0];
var jump = false;
var attack = false;

var player = {
  player: true,
  hp: 8,
  maxHp: 8,
  hitsBounds: true,
  hitsSolidObjects: true,
  graphic: 'player1',
  pos: [150, 55],
  size: [10, 20],
  vel: [0, 0],
  dir: 1,
  airborne: true,
  floor: null
};

var objects = [player];

var vadd = MXD.Vector.add;

function addObject(object) {
  objects.push(object);
  return object;
}

function prependObject(object) {
  objects.unshift(object);
  return object;
}

function removeDeadObjects() {
  _.remove(objects, {dead: true}).forEach(function(object) {
    if (object.onDie)
      object.onDie(object);
  });
}

function type(name) {
  return _.filter(objects, function(o) { return o[name]; });
}

function vaddi(a, b) {
  a[0] += b[0];
  a[1] += b[1];
}

function vaddScalei(a, b, f) {
  a[0] += f * b[0];
  a[1] += f * b[1];
}

function key(e, state) {
  var sym;
  if (e.keyCode == 'A'.charCodeAt(0)) sym = 'left';
  else if (e.keyCode == 'D'.charCodeAt(0)) sym = 'right';
  else if (e.keyCode == 'J'.charCodeAt(0)) sym = 'jump';
  else if (e.keyCode == 'K'.charCodeAt(0)) sym = 'attack';
  else return;

  keys[sym] = state;

  intent = [0, 0];
  if (keys.left == 'down') vaddi(intent, [-1, 0]);
  if (keys.right == 'down') vaddi(intent, [1, 0]);

  jump = (keys.jump == 'down');
  attack = (keys.attack == 'down');
}

function clear() {
  context.clearRect(0, 0, canvas.width, canvas.height);
}

function getAttackBox() {
  var hw = 10;
  return Rect.fromSize(
    V.add(V.add(player.pos, V.add([5, 9], [-hw, 0])), V.scale([10, 0], player.dir)),
    [2*hw, 6]
  );
}

function drawHUD() {
  for (var i = 0; i < player.maxHp / 2; i++) {

    var gx;
    if (i * 2 + 1 < player.hp)
      gx = gfx.heart;
    else if (i * 2 < player.hp)
      gx = gfx['heart-half'];
    else
      gx = gfx['heart-off'];
    context.drawImage(gx, 25 * i + 5, 5);
  }
}

var bricksCanvas;
var bricksContext;

function drawBricks() {

  context.drawImage(gfx.background, 0, 0);
  if (!bricksCanvas) {
    bricksCanvas = document.createElement('canvas');
    bricksContext = bricksCanvas.getContext('2d');

    for (var i=0; i<11; i++)
      for (var j=0; j<15; j++) {

        var x = i * 30 + ((j % 2 == 0) ? -15 : 0);
        var y = j * 10;
        bricksContext.drawImage(gfx.brick, x, y);
      }

  }

  context.drawImage(bricksCanvas, 0, 0);
}

function removeBricksNear(pos, radius) {
  var removed = [];
  for (var i=0; i<11; i++)
    for (var j=0; j<15; j++) {

      var x = i * 30 + ((j % 2 == 0) ? -15 : 0);
      var y = j * 10;

      var center = [x + 15, y + 5];
      if (V.magSq(V.sub(center, pos)) < radius * radius) {
        bricksContext.clearRect(x, y, 30, 10);

        var v = [
          Math.random() * 80 - 40,
          -80
        ];

        prependObject({
          pos: [x, y],
          size: [30, 10],
          graphic: 'brick',
          vel: v,
          gravity: 2
        });
      }
    }
}

function drawObject(object) {

  if (!object.color && !object.graphic)
    return;

  if (object.recovery && frameNo % 2 == 0)
    return;

  if (object.color) {
    context.fillStyle = object.color;
    context.fillRect(
      Math.floor(object.pos[0]),
      Math.floor(object.pos[1]),
      object.size[0],
      object.size[1]
    );
  }
  else if (object.graphic || object.anim) {

    var graphicName;
    if (object.anim) {
      var frame = Math.floor(frameNo / 6) % object.anim.length;
      graphicName = object.anim[frame];
    }
    else
      graphicName = object.graphic;


    var image = gfx[graphicName];

    var tl;
    if (object.aligntl) {
      tl = object.pos;
    }
    else  {
      var c = V.add(object.pos, V.scale(object.size, .5));
      tl = V.sub(c, V.scale([image.width, image.height], .5));
    }

    context.save();
    context.translate(
      Math.floor(tl[0]),
      Math.floor(tl[1])
    );
    if (object.flip) {
      context.translate(image.width, 0);
      context.scale(-1, 1);
    }
    context.drawImage(
      image, 0, 0
    );
    context.restore();
  }
}

function drawPlayer() {
  drawObject(player);

}

function collideObjectWithBounds(object) {
  if (object.pos[1] > floor - object.size[1]) {
    object.pos[1] = floor - object.size[1];
    object.vel[1] = 0;
    object.airborne = false;
    object.hitBounds = true;
  }

  if (object.pos[0] < 0) {
    object.pos[0] = 0;
    object.vel[0] = 0;
    object.hitBounds = true;
    object.hitWall = true;
  }

  if (object.pos[0] > 300 - object.size[0]) {
    object.pos[0] = 300 - object.size[0];
    object.vel[0] = 0;
    object.hitBounds = true;
    object.hitWall = true;
  }
}

function collide(a, b) {
  var ra = MXD.Rect.fromSize(a.pos, a.size);
  var rb = MXD.Rect.fromSize(b.pos, b.size);

  ra.axes = a.axes;
  rb.axes = b.axes;
  return collideRectRect(ra, rb);
}

function collideObjectWithSolidObject(object, platform) {
  var collision = collide(platform, object);
  if (collision) {
    object.pos = V.add(object.pos, V.scale(collision.axis, collision.depth));

    var tangent = V.perpendicular(collision.axis);
    if (collision.axis[1] < 0) {
      object.airborne = false;
      object.floor = platform;
    }

    if (collision.axis[1] == 0) {
      object.hitWall = true;
      debug('hw', Math.random());
    }

    object.vel = V.scale(tangent, V.dot(tangent, object.vel));
    object.hitObject = platform;
    if (!object.player) //lol
      platform.hitObject = 1;
  }
}

function collidePlayerWithEnemy(player, enemy) {
  if (player.recovery)
    return;
  if (collide(player, enemy)) {
    if (enemy.projectile)
      enemy.dead = true;
    player.hp -= 1;
    if (player.hp) {
      sfx.hit();
      player.recovery = 0.5;

      var diff = V.sub(player.pos, enemy.pos);
      diff[1] = -4;

      player.vel = V.add(player.vel, V.scale(V.norm(diff), 125));
    }
    else {
      paused = true;
      addObject({
        pos: [0, 0],
        size: [300, 150],
        graphic: 'gameover'
      });
      window.addEventListener('keydown', function() {
        window.location.reload();
      });
    }
  }
}

function collideAttackWithObject(attackBox, enemy) {
  var enemyRect = MXD.Rect.fromSize(enemy.pos, enemy.size);
  if (collideRectRect(attackBox, enemyRect)) {
    enemy.attacked = true;

    if (enemy.hp && !enemy.recovery) {
      enemy.hp -= 1;
      enemy.recovery = 0.2;
      if (enemy.hp == 0) {
        enemy.dead = true;
        sfx[enemy.dieSound || 'bang']();
      }
      else
        sfx[enemy.hitSound || 'big-hit']();
    }


  }
}


var debugData = {};
function debug(k, v) {
  debugData[k] = v;
  var e = document.getElementById('debug');
  if (!e) {
    e = document.createElement('pre');
    e.id = 'debug';
    document.body.appendChild(e);
  e.style.display = 'none';
  }
  e.textContent = JSON.stringify(debugData, null, 2);
}

function updatePlayer(dt) {

  var useIntent = (player.attack && !player.airborne) ? [0, 0] : intent;

  var floorVel = player.floor && player.floor.vel || [0, 0];
  var rvel = useIntent[0] * walkSpeed + floorVel[0];

  if (useIntent[0]) {
    if (!player.attack) {
      player.anim = ['player1', 'player2', 'player1', 'player3'];
    }
    player.flip = (useIntent[0] < 0);
  }
  else
    delete player.anim;

  var force = vadd(
    [(rvel - player.vel[0]) * friction, 0],
    gravity
  );

  vaddScalei(player.vel, force, dt);

  if (player.attackTimer) {
    player.attackTimer -= dt;
    if (player.attackTimer <= 0) {

      if (player.attack == 1) {
        player.attack = 2;
        player.attackTimer = attackTime;
        delete player.anim;
        player.graphic = 'player5';
        sfx.attack();
      }
      else {
        delete player.attack;
        delete player.attackTimer;
        player.graphic = 'player1';
      }

    }
  }
}

function moveEnemy(enemy, dt) {
  vaddScalei(enemy.vel, gravity, dt);
  vaddScalei(enemy.pos, enemy.vel, dt);

  enemy.flip = (enemy.vel[0] < 0);
}

function processPlayerControls() {
  if (jump && !player.airborne) {
    player.airborne = true;
    player.vel[1] = -jumpVel;
    sfx.jump();
  }

  if (intent[0] > 0)
    player.dir = 1;
  if (intent[0] < 0)
    player.dir = -1;

  if (attack && !player.attack) {
    player.attack = 1;
    player.attackTimer = preAttackTime;
    player.graphic = 'player4';
    delete player.anim;
  }
}

function processEnemyAI(enemy) {

  if (enemy.attacked)
    enemy.dead = true;

  if (enemy.hitWall) {
    enemy.hitWall = false;
    enemy.dir *= -1;
  }
  if (!enemy.airborne) {
    enemy.vel[0] = enemy.dir * 20 + (enemy.floor && enemy.floor.vel || [0, 0])[0];
  }
  else
    enemy.vel[0] = 0;
}

function updateMovingPlatform(platform, dt) {

  if (platform.triggered) {
    platform.enabled = !platform.enabled;
    platform.triggered = false;
  }

  if (platform.hitBounds || platform.hitObject)
    platform.dir *= -1;

  if (platform.enabled) {
    platform.vel[0] = platform.dir * platform.speed;
  }
  else
    platform.vel[0] = 0;
}

function processTriggers() {

  for (var trigger of type('trigger')) {
    if (trigger.test()) {
      trigger.effect();
      trigger.dead = true;
    }
  }
}

function collideObjects(a, b) {
  if (a.solid && b.hitsSolidObjects && !b.fallTo)
    collideObjectWithSolidObject(b, a);

  else if (a.player && (b.enemy || b.deadly))
    collidePlayerWithEnemy(a, b);

  else
    return false;

  return true;
}

function doCollision() {

  var n = objects.length;

  for (var i = 0; i < n; i++) {
    var object = objects[i];

    if (object.pos && object.size) {
      object.floor = null;
      object.airborne = true;
      object.hitBounds = false;
      object.hitWall = false;
      object.hitObject = false;

      if (object.hitsBounds)
        collideObjectWithBounds(object);

      for (var j = 0; j < i; j ++) {
        var other = objects[j];

        if (other.pos && other.size)
          collideObjects(object, other) || collideObjects(other, object);
      }
    }
  }

  var attackBox = (player.attack == 2) && getAttackBox();
  for (var object of objects) {
    if (object.target) {
      object.attacked = false;
      if (attackBox)
        collideAttackWithObject(attackBox, object);
    }
  }
}

function doPhysics(dt) {
  for (var object of objects) {

    if (object.gravity)
      vaddScalei(object.vel, gravity, dt * object.gravity);

    if (object.pos && object.vel) {
      vaddScalei(object.pos, object.vel, dt);
    }
  }
}

function update(dt) {
  processPlayerControls();
  updatePlayer(dt);
  doPhysics(dt);

  for (var platform of type('movingPlatform')) {
    updateMovingPlatform(platform, dt);
  }

  for (var enemy of type('enemy')) {
    processEnemyAI(enemy);
    moveEnemy(enemy, dt);
  }

  for (var object of objects) {

    if (object.recovery) {
      object.recovery -= dt;
      if (object.recovery <= 0)
        delete object.recovery;
    }

    if (object.fallTo) {
      if (object.pos[1] < object.fallTo) {
        object.vel[1] += gravity[1] * dt;
      }
      else {
        object.vel[1] = 0;
        if (object.onFall)
          object.onFall(object);
        delete object.fallTo;
      }
    }
  }

  doCollision();
  removeDeadObjects();

  for (var object of objects) {
    if (object.onAttack && object.attacked) {
      object.onAttack(object);
      object.attacked = false;
    }

    if (object.update)
      object.update(object, dt);

    if (object.ttl) {
      object.ttl -= dt;
      if (object.ttl <= 0) {
        delete object.ttl;
        object.dead = true;
      }
    }

    if (object.drag) {
      var d = V.scale(object.vel, -object.drag);
      object.vel = V.add(object.vel, d * dt);
    }
  }

  processTriggers();
  removeDeadObjects();

  debug('nobj', objects.length);
}

function runLoop(update) {
  var lastTime = 0;
  function nextFrame() {
    window.requestAnimationFrame(frame);
  }
  function frame(time) {
    var dt = time - lastTime;
    lastTime = time;
    if (!paused) {

      var fdt = dt / 1000;
      if (fdt > (1 / 60))
        fdt = 1 / 60;

      update(fdt);
      clear();
      drawBricks();

      drawPlayer();

      for (var object of objects)
        drawObject(object);

      drawHUD();
    }
    frameNo ++;
    nextFrame();
  }
  nextFrame();
}
